#######################################################################
#                             Parameters                              #
#######################################################################


abstract type AbstractParameters_Evolution end


@option "Evolution" struct Parameters_Evolution <: dg1d.AbstractParameters_Evolution
  "Time stepping algorithm, choose one out of LSERK, RK4, SSPRK3, ..."
  algorithm::String = "LSERK"
  "Start time of evolution"
  tstart::Float64 = 0
  "End time of evolution"
  tend::Float64 = 1
end


function sanitize(prms::Parameters_Evolution)
  @assert prms.algorithm in [ "LSERK", "RK4", "SSPRK3" ]
  @assert tstart < tend
end


#######################################################################
#                          Butcher-Tablaeus                           #
#######################################################################


@doc """
2nd order explicit Runge Kutta method -- midpoint formula.
"""
Base.@kwdef struct BT_MIDPT
  nstages = 2
  a = [ 0.0     0.0;
        0.5     0.0 ]
  b = [ 0.0,    1.0 ]
  c = [ 0.0,    0.5 ]
end


@doc """
4th order explicit Runge Kutta method.
"""
Base.@kwdef struct BT_RK4
  nstages = 4
  a = [ 0.0     0.0     0.0     0.0;
        0.5     0.0     0.0     0.0;
        0.0     0.5     0.0     0.0;
        0.0     0.0     1.0     0.0 ]
  b = [ 1.0/6.0,  1.0/3.0,  1.0/3.0,  1.0/6.0 ]
  c = [ 0.0,      0.5,      0.5,      1.0 ]
end


@doc """
4th order explicit Runge Kutta method -- two thirds version.
"""
Base.@kwdef struct BT_RK4_twothirds
  nstages = 4
  a = [ 0.0       0.0       0.0     0.0;
        1.0/3.0   0.0       0.0     0.0;
        -1.0/3.0  1.0       0.0     0.0;
        1.0       -1.0      1.0     0.0 ]
  b = [ 1.0/8.0,  3.0/8.0, 3.0/8.0, 1.0/8.0 ]
  c = [ 0.0,      1.0/3.0, 2.0/3.0, 1.0 ]
end


@doc """
4th order explicit Runge Kutta method -- Ralston's method,
designed to have minimal truncation error.
cf. https://en.wikipedia.org/wiki/List_of_Runge%E2%80%93Kutta_methods#Classic_fourth-order_method
"""
Base.@kwdef struct BT_RK4_Ralston
  nstages = 4
  a = [ 0.0         0.0           0.0         0.0;
        0.4         0.0           0.0         0.0;
        0.29697761  0.15875964    0.0         0.0;
        0.21810040  -3.05096516   3.83286476  0.0 ]
  b = [ 0.17476028, -0.55148066,  1.20553560, 0.17118478  ]
  c = [ 0.0,        0.4,          0.45573725, 1.0         ]
end


@doc """
Strong stability preserving Runge Kutta 3rd order method.
"""
Base.@kwdef struct BT_SSPRK3
  nstages = 3
  a = [ 0.0     0.0     0.0;
        1.0     0.0     0.0;
        0.25    0.25    0.0 ]
  b = [ 1.0/6.0,  1.0/6.0,  2.0/3.0 ]
  c = [ 0.0,      1.0,      0.5 ]
end


@doc """
Runge-Kutta-Fehlberg 4/5 method,
cf. https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta%E2%80%93Fehlberg_method
"""
@with_kw struct BT_RKF45
  nstages = 6
  a = [ 0           0           0           0           0       0;
        1/4         0           0           0           0       0;
        3/32        9/32        0           0           0       0;
        1932/2197   -7200/2197  7296/2197   0           0       0;
        439/216     -8          3680/513    -845/4104   0       0;
        -8/27       2           -3544/2565  1859/4104   -11/40  0 ]
  b1 = [ 16/135     0           6656/12825  28561/56430 -9/50   2/55 ]
  b2 = [ 25/216     0           1408/2565   2197/4104   -1/5    0 ]
  c  = [ 0          1/4         3/8         12/13       1       1/2 ]
end


@doc """
Low storage explicit Runge Kutta 4th order method.
Taken from Hesthaven, Warburton (2007), section 3.4.
"""
Base.@kwdef struct COEFFS_LSERK4
  nstages = 5
  a = [            0.0,
          -567301805773.0/1357537059087.0,
          -2404267990393.0/2016746695238.0,
          -3550918686646.0/2091501179385.0 ,
          -1275806237668.0/842570457699.0]
  b = [ 1432997174477.0/9575080441755.0,
           5161836677717.0/13612068292357.0,
           1720146321549.0/2090206949498.0,
           3134564353537.0/4481467310338.0,
           2277821191437.0/14882151754819.0]
  c = [             0.0,
           1432997174477.0/9575080441755.0,
           2526269341429.0/6820363962896.0,
           2006345519317.0/3224310063776.0,
           2802321613138.0/2924317926251.0]
end


#######################################################################
#                              Steppers                               #
#######################################################################


@doc """
Stepper for Low Storage five-stage fourth-order Explicit Runge Kutta scheme.
Taken from Hesthaven, Warburton (2007), section 3.4.
u! ... mutable RHS of ODE
F ... RHS functional
u0 ... initial data
t ... current time
dt ... time step
coeffs ... Explicit Runge Kutta coefficients
"""
function LSERK4_step!(u!::tV, F, u0::tV, t::tS, dt::tS, tmp_k, coeffs) where {tV, tS}
  @unpack nstages, a, b, c = coeffs

  u! .= u0
  kim1, ki = tmp_k

  for i in 1:nstages

    kim1 .= ki
    F(ki, u!, t + c[i] * dt)
    ki *= dt
    if i > 1
      ki .+= a[i] * kim1
    end
    u! .+= b[i] * ki

  end # for

  return
end


@doc """
Base stepper for Explicit Runge Kutta scheme.
u! ... mutable RHS of ODE
F ... RHS functional
u0 ... initial data
t ... current time
dt ... time step
coeffs ... Explicit Runge Kutta coefficients, e.g. Butcher tableau
"""
function baseERK_step!(u!::tV, F, u0::tV, t::tS, dt::tS, tmp_k, coeffs) where {tV, tS}
  @unpack nstages, a, b, c = coeffs

  for n in 1:nstages
    tdt = t + c[n] * dt
    u! .= u0
    for i in 1:n-1
      u! .+= dt * a[n, i] * tmp_k[i]
    end # for
    F(tmp_k[n], u!, tdt)
  end # for

  u! .= u0
  for n in 1:nstages
    u! .+= dt * b[n] * tmp_k[n]
  end

  return
end


@doc """
Intended to use with baseERK_step and LSERK4_step as temporary storage for
RK substeps.
size ... size of state vector
n ... number of stages
"""
function make_RK_stages(size, n)
  return [ zeros(size) for _ in 1:n ]
end


@doc """
u0 ... initial data
scheme ... Explicit Runge Kutta scheme
TODO Replace scheme::Symbol with an @enum
"""
function make_ERK_stepper(u0, scheme::Symbol)

  coeffs = nothing
  method = nothing
  if      scheme == :MIDPT
    coeffs = BT_MIDPT()
    method = baseERK_step!

  elseif  scheme == :RK4
    coeffs = BT_RK4()
    method = baseERK_step!

  elseif  scheme == :RK4_twothirds
    coeffs = BT_RK4_twothirds()
    method = baseERK_step!

  elseif  scheme == :RK4_Ralston
    coeffs = BT_RK4_Ralston()
    method = baseERK_step!

  elseif  scheme == :SSPRK3
    coeffs = BT_SSPRK3()
    method = baseERK_step!

  elseif  scheme == :LSERK4
    coeffs = COEFFS_LSERK4()
    method = LSERK4_step!

  else
    error("Requested ERK scheme not implemented!")
  end # if

  tmp_k = make_RK_stages(size(u0), coeffs.nstages)

  stepper(u!, F, u0, t, dt) = method(u!, F, u0, t, dt, tmp_k, coeffs)

  return stepper
end



isinvalid(u) = any(isnan.(u)) || any(isinf.(u))


@doc """
Explicit Runge Kutta solver.
F ... F(u!, u0, t), RHS of ODE
      u! ... mutable RHS
      u0 ... initial state
      t ... time
u0 ... initial data
dt ... time step
tspan ... (t_start, t_end), time interval
scheme ... explicit Runge Kutta scheme
post_hook(u, t) ... TODO Add signature here
"""
function evolve_ERK(
    F::Function, u0::AbstractArray,
    tstep::Union{Function,Float64}, tspan::Tuple{Real,Real}, scheme::Symbol;
    post_hook::Union{Function,Nothing}=nothing)

  un = deepcopy(u0); unp1 = similar(un)
  stepper! = make_ERK_stepper(u0, scheme)

  tstart, tend = tspan
  tn = tstart
  prog = Progress(100) # 100 %
  while tn < (tend-tstart)

    prcnt = tn / (tend-tstart) * 100
    ProgressMeter.update!(prog, trunc(Int,prcnt))

    if isa(tstep, Function)
      dt = tstep(un, tn)
    else
      dt = tstep
    end

    if tn + dt > tend
      dt = tend - tn
    end

    stepper!(unp1, F, un, tn, dt)
      if isinvalid(unp1)
      println()
      error("Nan or Inf encountered in evolution at t = $(tn)! Aborting ...")
    end

    un .= unp1
    tn += dt

    if tn >= tend
      ProgressMeter.finish!(prog)
    end

    if !isnothing(post_hook)
      success::Bool = post_hook(un, tn)
      if !success
        error("post_hook failed! Aborting ...")
      end
    end
  end # while

  return unp1
end



@doc """
Adapative and explicit Runge Kutta solver.
F ... F(u!, u0, t), RHS of ODE
      u! ... mutable RHS
      u0 ... initial state
      t ... time
u0 ... initial data
dt ... time step
reltol ... max relative tolerance in error between consecutive time steps
abstol ... max absolute tolerance in error between consecutive time steps
tspan ... (t_start, t_end), time interval
scheme ... adaptive explicit Runge Kutta scheme
post_hook ... TODO Add signature here
"""
function evolve_adaptive_ERK(F, u0, tstep, tspan; reltol=1e-10, abstol=1e-10,
    scheme::Symbol=:RK4, post_hook::Union{Function,Nothing}=nothing)

  un = deepcopy(u0); unp1 = similar(un);
  tmp_u1 = similar(unp1); tmp_u2 = similar(unp1);
  stepper! = make_ERK_stepper(u0, scheme)

  adaptive_max_its = 1000 # maximum number of adaptive iterations

  tstart, tend = tspan
  tn = tstart
  prog = Progress(100) # 100 %
  while tn < (tend-tstart)

    prcnt = tn / (tend-tstart) * 100
    ProgressMeter.update!(prog, trunc(Int,prcnt))

    if isa(tstep, Function)
      dt = tstep(un, tn)
    else
      dt = tstep
    end

    if tn + dt > tend
      dt = tend - tn
    end

    adapted_dt = dt
    norm_un = norm(un, 1)
    # estimate error by difference between a single RK step and
    # two consecutive RK step evolution
    # iterate this until tolerance is reached
    while true
      stepper!(unp1, F, un, tn, adapted_dt)
      stepper!(tmp_u1, F, un, tn, adapted_dt*0.5)
      stepper!(tmp_u2, F, tmp_u1, tn+adapted_dt*0.5, adapted_dt*0.5)

      if isinvalid(unp1) || isinvalid(tmp_u1) || isinvalid(tmp_u2)
        println()
        error("Nan or Inf encountered in evolution at t = $(tn)! Aborting ...")
      end
      
      err = norm(unp1 .- tmp_u2, 1)
      if err <= reltol * norm_un + abstol
        break
      end
      adapted_dt *= 0.5
      if log(dt / adapted_dt)/log(2) > adaptive_max_its
        println()
        error("Could not refine solution after $adaptive_max_its iterations! Aborting ... ")
      end
    end

    un .= unp1
    tn += adapted_dt

    if tn >= tend
      ProgressMeter.finish!(prog)
    end

    if !isnothing(post_hook)
      success::Bool = post_hook(un, tn)
      if !success
        println()
        @info "Termination requestd by post_hook!"
        break
      end
    end
  end # while

  return unp1
end
