@doc """
  `Mesh`

Basis for modeling a Discontinuous Galerkin scheme in 1D.
"""
struct Mesh
  element::SpectralElement  # collection of spectral element operators
  K::Int64                  # number of cells
  xmin::Float64             # left most grid coordinate
  xmax::Float64             # right most grid coordiante
  CFL::Float64              # Courant Friedrichs Lewy factor
  L::Float64                # domain length
  dl::Float64               # equidistant subcell length
  x::Matrix{Float64}        # grid points
  invjac::Matrix{Float64}   # inverse jacobian of coordinate transformation
                            # from computational [-1,1] onto physical grid [xmin,xmax]

  function Mesh(element::SpectralElement, K::Int64, xmin, xmax, CFL)
    @unpack z = element
    xmin = float(xmin)
    xmax = float(xmax)
    CFL = float(CFL)
    L = xmax - xmin
    dl = L/K
    a = [ xmin + dl * idx for idx = 0:K-1 ]
    b = a .+ dl
    N = length(z)
    x = Matrix{Float64}(undef, N, K)
    invjac = similar(x)
    for i=1:N, j=1:K
      aj = a[j]
      bj = b[j]
      zi = z[i]
      x[i,j] = 0.5 * ((bj - aj) * zi + (aj + bj))
      invjac[i,j] = 2.0 / (bj - aj)
    end
    # sanitize cell end points to have same x coordinate
    for j=1:K-1
      xl = x[end,j]
      xr = x[1,j+1]
      xmean = 0.5 * (xl + xr)
      x[end,j] = xmean
      x[1,j+1] = xmean
    end
    return new(element, K, xmin, xmax, CFL, L, dl, x, invjac)
  end
end


#######################################################################
#                             Parameters                              #
#######################################################################


abstract type AbstractParameters_Mesh <: AbstractParameters end


@option "Mesh" struct Parameters_Mesh <: dg1d.AbstractParameters_Mesh
  "Polynomial order of spectral element"
  N::Int64 = 5
  "Number of spectral elements spanning domain"
  K::Int64 = 20
  "Courant-Friedrich-Lewy number"
  CFL::Float64 = 0.5
  "Quadrature rule underlying the spectral element, chose one of 'LGL, GLGL, LGL_lumped'."
  quadr::String = "LGL"
  "Minimum domain extension x direction"
  xmin::Float64 = -1.0
  "Maximum domain extension x direction"
  xmax::Float64 = 1.0
end


function sanitize(prms::Parameters_Mesh)
  @assert prms.N > 0 "Invalid polynomial order requested!"
  @assert prms.K > 0 "Invalid number of cells requested!"
  @assert prms.CFL > 0 "CFL number must be positive!"
  @assert prms.quadr in [ "LGL", "GLGL", "LGL_lumped" ] "Unsupported quadrature rule provided!"
  @assert prms.xmin < prms.xmax "Invalid domain size specified!"
end


function Mesh(prms::AbstractParameters_Mesh)
  @unpack N, quadr, K, xmin, xmax, CFL = prms
  element = SpectralElement(N, Symbol(quadr))
  mesh = Mesh(element, K, xmin, xmax, CFL)
  return mesh
end


#######################################################################
#                              Utilities                              #
#######################################################################


@doc """
  `grid_average(u, mesh::Mesh)`

Return the grid average of all cell averages 
"""
function grid_average(u, mesh::Mesh)
  @unpack element, dl, L = mesh
  cellavg = cellwise_average(u, mesh) .* dl / L
  avg = sum(cellavg)
  return avg
end


@doc """
  `cellwise_average(u, mesh::Mesh)`

Return the cellwise average of `u` over `mesh.x`.
"""
function cellwise_average(u, mesh::Mesh)
  @unpack element, invjac, K, dl = mesh
  @unpack Npts = element
  v = ones(Npts)
  avg = zeros(K)
  for i in 1:K
    avg[i] = element.iproduct(u[:,i], v)
  end
  avg ./= invjac[1,:] * dl
  return avg
end


@doc """
  `cellwise_inner_product(u, v, mesh::Mesh)`

Return vector of inner products between `u, v` on every cell in `mesh`.
"""
function cellwise_inner_product(u, v, mesh::Mesh)
  @unpack element, K, invjac = mesh
  @unpack Npts = element
  inner_products = [ element.iproduct(u[:,k], u[:,k]) for k = 1:K ]
  inner_products ./= invjac[1,:]
  return inner_products
end


@doc """
  `broken_inner_product(u, v, mesh::Mesh)`
  
Return sum of all inner products between `u,v` on every cell in `mesh`.
"""
function broken_inner_product(u, v, mesh::Mesh)
  inner_products = cellwise_inner_product(u, v, mesh)
  total = sum(inner_products[:])
  return total
end


@doc """
Compute local maximum absolute value.
"""
function local_maxabs(v::AbstractMatrix, mesh::Mesh) 
  return vec(maximum(abs.(v), dims=1))
end


@doc """
Compute local minimum absolute value.
"""
function local_maxabs(v::AbstractVector, mesh::Mesh) 
  return local_maxabs(reshape(v, (mesh.element.Npts, mesh.K)))
end


@doc """
Return index of coordinate in `mesh` that is closest to point `xi`.
"""
function locate_point(xi, mesh::Mesh)
  min, idx = LinearAlgebra.findmin( @. abs(mesh.x - xi) )
  return Tuple(idx)
end


#######################################################################
#                             DEPRECATED                              #
#######################################################################


@doc """

"""
@with_kw struct Mesh_Parameters <: dg1d.AbstractParameters
  N::Int64                  = 5
  K::Int64                  = 20
  CFL::Float64              = 0.2
  quadr::Symbol             = :LGL
  xmin::Float64             = -1.0
  xmax::Float64             = 1.0
end
@deprecate Mesh_Parameters Parameters_Mesh


@doc """
  `Mesh(prms::Mesh_Parameters)`

Constructor for `Mesh` type from `Mesh_Parameters`.
"""
function Mesh(prms::Mesh_Parameters)
  @unpack_Mesh_Parameters prms
  element = SpectralElement(N, quadr)
  mesh = Mesh(element, K, xmin, xmax, CFL)
  return mesh
end
