module EquationOfState

import dg1d: AbstractParameters, @option, @with_kw, @unpack

export Polytrope, Parameters_Polytrope,
       Derivative, Density, Pressure, Enthalpy, TotalEnergy, InternalEnergy


#######################################################################
#                              Interface                              #
#######################################################################


@doc """
Supertypes for equation of state variables.

# Example

```julia
rho_val = 1.23
rho = Density(rho_val) # promotion to equation of state variable
```

See `AbstractEquationOfState` on how to use it.
"""
abstract type AbstractVariable end
@inline (var::AbstractVariable)() = var.x
struct Derivative{T<:AbstractVariable, R<:AbstractVariable} end


@doc """
Equations of state (eos) must subtype `AbstractEquationOfState`.

Evaluation of state variables is done through

  (eos::AbstractEquationOfState)(::Type{AbstractVariable}, var1::AbstractVariable, var2::AbstractVariable, ...)
  (eos::AbstractEquationOfState)(::Type{Derivative{AbstractVariable, AbstractVariable}}, 
                                  var1::AbstractVariable, var2::AbstractVariable, ...)

# Example

```julia
eos = Poltrope()
p = eos(Pressure, Density(1.0))
dpdrho = eos(Derivative{Pressure, Density}, Density(rho))
```

Use `method(eos)` to query a list of available computations for a given equation of state `eos`.
"""
abstract type AbstractEquationOfState end


# supertype for parameter interface with EquationOfState model
abstract type AbstractParameters_EquationOfState end


### Error handling

function (eos::AbstractEquationOfState)(
    f::Type{<:AbstractVariable}, vars::AbstractVariable...)
  vars_str = join(vars, ", ")
  error("Computation of '$(f)' for equation of state " *
        "'$(typeof(eos))' with arguments '$(vars_str)' not implemented!")
end


function (eos::AbstractEquationOfState)(
    f::Type{Derivative{T,R}}, vars::AbstractVariable...) where {T<:AbstractVariable,R<:AbstractVariable}
  vars_str = join(vars, ", ")
  error("Computation of '$(f)' for equation of state " *
        "'$(typeof(eos))' with arguments '$(vars_str)' not implemented!")
end


#######################################################################
#                        Variable declerations                        #
#######################################################################

# convinience macro
macro Variable(varname, docstr="")
  esc(
    quote
      @doc $docstr
      struct $varname{T} <: EquationOfState.AbstractVariable x::T end
      export $varname
    end
  )
end


@Variable Density """
Denoted by ρ.
"""


@Variable Pressure """
Denoted by p.
"""


@Variable InternalEnergy """
Denoted by ϵ.
"""


@Variable TotalEnergy """
Denoted by e and defined as

  e = ρ (1 + ϵ).

"""

@Variable Enthalpy """
Denoted by h and defined as

  h = 1 + ϵ + p / ρ

"""

#######################################################################
#                           Implementations                           #
#######################################################################


include("polytropic_eos.jl")


end
