@with_kw struct Polytrope <: AbstractEquationOfState
  K::Float64 = 1.0
  Gamma::Float64 = 5.0/3.0
end


#######################################################################
#                             Parameters                              #
#######################################################################


@option "Polytrope" struct Parameters_Polytrope <: AbstractParameters_EquationOfState
  "Polytropic constant"
  K::Float64 = 100
  "Polytropic exponent"
  Gamma::Float64 = 2
end


function sanitize(prms::Parameters_Polytrope)
  @assert K > 0 "Polytropic constant must be positive!"
  @assert Gamma > 0 "Polytropic index must be positiv!"
end


# constructor based on parameters
function Polytrope(prms::Parameters_Polytrope)
  @unpack K, Gamma = prms
  return Polytrope(K, Gamma)
end


#######################################################################
#                            Computations                             #
#######################################################################


# p = p(ρ)
function (eos::Polytrope)(::Type{Pressure}, rho::Density)
  @unpack K, Gamma = eos
  return K * rho()^Gamma
end
  

# ρ = ρ(p)
function (eos::Polytrope)(::Type{Density}, p::Pressure)
  @unpack K, Gamma = eos
  return (p() / K)^(1/Gamma)
end


# h = h(p,ρ)
function (eos::Polytrope)(::Type{Enthalpy}, p::Pressure, rho::Density)
  @unpack K, Gamma = eos
  return 1 + Gamma / (Gamma - 1) * p() / rho()
end


# ϵ = ϵ(p, ρ)
function (eos::Polytrope)(::Type{InternalEnergy}, p::Pressure, rho::Density)
  @unpack K, Gamma = eos
  h = eos(Enthalpy, p, rho)
  return h - 1 - p() / rho()
end


# e = e(p, ρ)
function (eos::Polytrope)(::Type{TotalEnergy}, p::Pressure, rho::Density)
  @unpack K, Gamma = eos
  eps = eos(InternalEnergy, p, rho)
  return rho() * (1 + eps)
end


# ∂p/∂ρ(ρ)
function (eos::Polytrope)(::Type{Derivative{Pressure, Density}}, rho::Density)
  @unpack K, Gamma = eos
  return K * Gamma * rho()^(Gamma-1)
end
