module ShockSensor



@doc """
  `linear_ramp_function(s::T, k::T) where T::Real`

According to Glaubitz, Phd thesis, 2020, section 10.2.5.
"""
function linear_ramp_function(s::T, k::T) where T<:Real
  if s <= k
    return T(1)
  elseif k < s < T(1)
    return (1 - s) / (1 - k)
  else
    return T(0)
  end
end



@doc """
  `shock_sensor(ramp_fn::Function, indicator::AbstractArray{T}, m::T, M::T) where T<:Real`

Use ramping function `ramp_fn` to reduce indicator values to normalized values in [0,1].
`m, M` control lower and upper limits of inidicator.
`ramp_fn` should accept a sensing value `s` ∈ [0,1] and return a value in [0,1].
"""
function shock_sensor(ramp_fn::Function, indicator::AbstractArray{T}, m::T, M::T) where T<:Real

  # apply limits m, M and normalize
  norm_indicator = deepcopy(indicator)
  M = maximum(indicator)
  if M == 0.0
    M = 1.0
  end
  m = 0.0
  norm_indicator[ indicator .< m] .= m
  norm_indicator[ indicator .> M] .= M
  norm_indicator ./= M - m
  norm_indicator .-= m

  sensor = @. ramp_fn.(norm_indicator)

  return sensor
end



end
