export SpectralElement

Base.@kwdef struct SpectralElement
  N           # polynomial order
  Npts        # number of quadrature points = N + 1
  z           # quadrature points
  w           # quadrature weights
  bw          # barycentric weights
  D           # = D_ij = l'_j(z_i)    # Kopriva (2009), p. 81
  S           # = ∫ l_i(z) l'_j(z) dz, stiffness matrix
  invM        # = M^-1, inverse mass matrix
  MDM         # = (M^-1 D^T M) = M^-1 S^T
  l_lhs       # = l_i(-1) ... Lagrange polynomial at z = -1
  l_rhs       # = l_i(+1) ... Lagrange polynomial at z = +1
  Ml_lhs      # = M^-1 l(-1)
  Ml_rhs      # = M^-1 l(+1)
  V           # = P_(j-1)(z_i), Vandermonde matrix
  invV        # inv(V), inverse Vandermonde matrix
  VFV         # filtering matrix
  quadr       # :LGL, :LGL_mass_lumped, :GLGL
  gridtype    # = quadr
  iproduct    # iproduct(f1, f2), inner product functor
  L           # = (P_i, P_j), linear dependency of matrix of Legendre polynomials
  E           # = (l_i, P_j), expansion matrix
  T           # transfer matrix, mapping nodal to modal coefficients
  invT        # inverse transfer matrix

  function SpectralElement(N::Int, quadr::Symbol)
    Npts = N + 1

    gridtype = quadr

    L = zeros(Npts, Npts)
    E = zeros(Npts, Npts)
    T = zeros(Npts, Npts)
    invT = zeros(Npts, Npts)

    if quadr === :LGL
      z, w        = LGL.rule(Npts)
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      S           = LGL.stiffness_matrix(Npts)
      invM        = LGL.inverse_mass_matrix_exact(Npts)
      # TODO Figure out how much the aliasing error influences the accuracy of the inner product.
      # Could be totally dealised using the Orszag 2/3rds rule.
      iproduct    = (f1,f2) -> LGL.inner_product(z, w, f1 .* f2)
      L           = LGL.linear_dependency_matrix_exact(z, w)
      E           = LGL.expansion_matrix_exact(z, w)
    elseif quadr === :LGL_lumped
      z, w        = LGL.rule(Npts)
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      S           = LGL.stiffness_matrix(Npts)
      invM        = LGL.inverse_mass_matrix_lumped(Npts)
      iproduct    = (f1,f2) -> LGL.inner_product(z, w, f1 .* f2)
      L           = LGL.linear_dependency_matrix_lumped(z, w)
      E           = LGL.expansion_matrix_lumped(z, w)
    elseif quadr === :GLGL
      z, w, v     = GLGL.rule(Npts)
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      D2          = second_derivative_matrix(z)
      M           = GLGL.mass_matrix(z, w, v, D)
      S           = GLGL.stiffness_matrix(z, w, v, D, D2)
      invM        = inv(M)
      iproduct    = (f1,f2) -> GLGL.inner_product(z, w, v, D, f1, f2)
      L           = GLGL.linear_dependency_matrix(z, w, v, D)
      E           = GLGL.expansion_matrix(z, w, v, D)
    else
      error("Invalid grid type specified! Use :LGL_lumped, :LGL, :GLGL.")
    end

    MDM     = invM * transpose(S)
    purge_zeros!(MDM)
    V       = vandermonde_matrix(z)
    invV    = inv(V)
    F       = filter_matrix(z)
    VFV     = V * F * invV
    T       = inv(transpose(E)) * transpose(L)
    invT    = inv(T)

    l_lhs   = zeros(N+1)
    l_rhs   = zeros(N+1)
    l_lhs[1]    = 1.0
    l_rhs[end]  = 1.0
    Ml_lhs = invM * l_lhs
    Ml_rhs = Ml_lhs[end:-1:1] # to ensure exact cancellation for continuous functions

    return new(N, Npts, z, w, bw, D, S, invM, MDM, l_lhs, l_rhs, Ml_lhs, Ml_rhs, 
               V, invV, VFV, quadr, gridtype, iproduct, L, E, T, invT)
  end
end
